// GA.C
// Genetic Algorithm, mapper problem
// John Dunkelberg, 1994

#include<stdlib.h>
#include<iostream.h>
#include<fstream.h>

#ifndef DEBUG
#define DEBUG 0
#else
#define DEBUG 1
#endif

#include "ga.H"
#include "meshmap.H"
#include "genemap.H"

extern "C"
{
  long random(void);
  void srandom(int);
  double drand48(void);
  void srand48(long);
  char *sprintf(const char *fmt, ... );
}

main(int argc,char **argv)
{
  int MAX_POP,MAX_ERA;
  long gene_count;
  float fitness,best_fit,best_of_era;
  double de_jong_online, de_jong_offline;
  int count;
  char *outname;

  if (argc != 5)
    {
      cout << "not right arguments" << endl;
      cout << "maxpop maxera meshfile seed" << endl;
      exit(1);
    }

  if (DEBUG)
    cout << "\nDebug level: "<<DEBUG<<endl;

  MAX_POP = atoi(argv[1]);
  MAX_ERA = atoi(argv[2]);
  srand48(atol(argv[4]));
  srandom(atol(argv[4]));

  if (DEBUG)
    {
      cout<<argv[0]<<" "<<argv[1]<<" "<<argv[2]<<" "<<argv[3]<<" ";
      cout<<argv[4]<<" "<<endl;
    }

  fstream meshfile(argv[3],ios::in);

  outname = new char[80];
  sprintf(outname,"./res.%d.%d.%s.%d",
	  MAX_POP,MAX_ERA,argv[3],atoi(argv[4]));

  fstream outfile(outname,ios::out);

  if (meshfile.bad())
    {
      cout << "Cannot open file" << argv[3] << endl;
      exit(1);
    }

  if (DEBUG)
    {
      cout << "opened file "<<argv[3]<<endl;
      cout << "opened file "<<outname<<endl;
    }
  Meshmap mesh(meshfile);

  if (DEBUG)
    {
      cout << "constructed mesh\n";
      mesh.Info();
    }

  Genemap alpha[MAX_POP],beta[MAX_POP];

  if (DEBUG)
    cout << endl << "Gene created" << endl;

  gene_count =0;
  de_jong_online = de_jong_offline = best_fit = best_of_era = 0.0;

  for(count=0;(count<MAX_POP);count++)
    {
      alpha[count].Randomize(&mesh);
      alpha[count].Mutate();
      alpha[count].Fitness();
      //alpha[count].Info();
      gene_count++;
      fitness = alpha[count].Get_Fit();
      if (fitness > best_fit)
	best_fit = fitness;
      de_jong_online += (double) fitness;
      de_jong_offline += (double) best_fit;
    }

  int a,b,era;
  float aval,bval;
  float *wheel;
  for(era=0;(era<MAX_ERA);era++)
    {
      // Create the wheel
      wheel = new float[MAX_POP];
      float sum,weight,prod;
      //cout << "\n\nWheel";
      for(count=0,sum=0;(count<MAX_POP);count++)
	{
	  for(weight=WHEEL_WEIGHT,prod=1;(weight>0);weight--)
	    prod *= alpha[count].Get_Fit();
	  sum += prod;
	  wheel[count] = sum;
	  //cout << " " << wheel[count];
	}
      //cout << endl;

      int which_best;
      for(count=0,which_best=-1,best_of_era=0.0;(count<MAX_POP);count++)
	{
	  aval = ((float) drand48()) * sum;
	  bval = ((float) drand48()) * sum;
	  
	  int in_cnt;
	  for(in_cnt=0;(wheel[in_cnt]<aval);in_cnt++);
	  a = in_cnt;
	  for(in_cnt=0;(wheel[in_cnt]<bval);in_cnt++);
	  b = in_cnt;

	  //cout << "\nCrossing "<<a<<"("<<aval<<") and "<<b<<"("<<bval<<")\n";
	  beta[count].Cross(alpha[a], alpha[b], &mesh);
	  beta[count].Mutate();
	  beta[count].Fitness();
	  
	  beta[count].Info();

	  gene_count++;
	  fitness = alpha[count].Get_Fit();
	  if (fitness > best_fit)
	      best_fit = fitness;
	  if (fitness > best_of_era)
	    {
	      best_of_era = fitness;
	      which_best = count;
	    }
	  de_jong_online += (double) fitness;
	  de_jong_offline += (double) best_fit;

	  if (!(gene_count % (MAX_POP/5)))
	    {
	      outfile<<era<<"."<<count;
	      outfile<<" "<<(de_jong_online/ (double) gene_count);
	      outfile<<" "<<(de_jong_offline/ (double) gene_count)<<endl;
	    }
	}
      delete wheel;

      if ((DEBUG) && (which_best != -1))
	{
	  cout << "\nBest in era "<<era<<":"<<which_best;
	  beta[which_best].Info();
	}

      for(count=0;(count<MAX_POP);count++)
	alpha[count] = beta[count];
    }

  outfile<<MAX_ERA<<".00";
  outfile<<" "<<(de_jong_online/ (double) gene_count);
  outfile<<" "<<(de_jong_offline/ (double) gene_count)<<endl;

  if (DEBUG)
    cout << "End of Run" << endl;
}
