// gene.H
// Gene header file for the Bokhari problem
// a mapping from the problem to the mesh
// John Dunkelberg

// Note: also defines the << operator on the Gene

class Gene
{
protected:
  int size;
  int edges;
  double fitness;
  int *map;
  
public:
  Gene(void);
  virtual void Set_Fit(float);
  virtual int Get_Map(int) const;        // get a particular mapping
  virtual inline double Get_Fit(void) const
    { return fitness; }          // return the fitness
  virtual inline int Get_Size(void) const
    { return size; }             // get the size of the mesh (nodes)
  virtual void Info();
};

// Constructs a Gene
Gene::Gene()
{
  size = 0;
  edges = 0;
  map = (int *) NULL;
  fitness = -1.0;
}

// Set the fitness value of the mapping gene
void Gene::Set_Fit(float newfit)
{
  fitness = newfit;
}

// Get a particular mapping... a bit too large for an inline
int Gene::Get_Map(int x) const
{
  if ((x >=0) && (x < size))
    return(map[x]);
  else
    return(-1);        // no such machine node to map
}

// Primarily for debugging, prints the gene mapping
void Gene::Info()
{
  int count;
  cout << endl << "S:"<< size << " F:" << fitness << " E:" << edges;
  for(count=0, cout << " G:";(count<size);count++)
    cout << count << "(" << map[count] << ") ";
  cout << endl;
}
